/* Copyright (C) 2010-2011 Linaro Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/*
 * Authors:
 *  Marc Ordinas i Llopis, Collabora Ltd <marc.ordinasillopis@linaro.org>
 */

#include "EGL/egl_proxy.h"

PFNPROXYEGLGETERROR                      pfnProxyEglGetError = NULL;
PFNPROXYEGLGETDISPLAY                    pfnProxyEglGetDisplay = NULL;
PFNPROXYEGLINITIALIZE                    pfnProxyEglInitialize = NULL;
PFNPROXYEGLTERMINATE                     pfnProxyEglTerminate = NULL;
PFNPROXYEGLQUERYSTRING                   pfnProxyEglQueryString = NULL;
PFNPROXYEGLGETCONFIGS                    pfnProxyEglGetConfigs = NULL;
PFNPROXYEGLCHOOSECONFIG                  pfnProxyEglChooseConfig = NULL;
PFNPROXYEGLGETCONFIGATTRIB               pfnProxyEglGetConfigAttrib = NULL;
PFNPROXYEGLCREATEWINDOWSURFACE           pfnProxyEglCreateWindowSurface = NULL;
PFNPROXYEGLCREATEPBUFFERSURFACE          pfnProxyEglCreatePbufferSurface = NULL;
PFNPROXYEGLCREATEPIXMAPSURFACE           pfnProxyEglCreatePixmapSurface = NULL;
PFNPROXYEGLDESTROYSURFACE                pfnProxyEglDestroySurface = NULL;
PFNPROXYEGLQUERYSURFACE                  pfnProxyEglQuerySurface = NULL;
PFNPROXYEGLBINDAPI                       pfnProxyEglBindAPI = NULL;
PFNPROXYEGLQUERYAPI                      pfnProxyEglQueryAPI = NULL;
PFNPROXYEGLWAITCLIENT                    pfnProxyEglWaitClient = NULL;
PFNPROXYEGLRELEASETHREAD                 pfnProxyEglReleaseThread = NULL;
PFNPROXYEGLCREATEPBUFFERFROMCLIENTBUFFER pfnProxyEglCreatePbufferFromClientBuffer = NULL;
PFNPROXYEGLSURFACEATTRIB                 pfnProxyEglSurfaceAttrib = NULL;
PFNPROXYEGLBINDTEXIMAGE                  pfnProxyEglBindTexImage = NULL;
PFNPROXYEGLRELEASETEXIMAGE               pfnProxyEglReleaseTexImage = NULL;
PFNPROXYEGLSWAPINTERVAL                  pfnProxyEglSwapInterval = NULL;
PFNPROXYEGLCREATECONTEXT                 pfnProxyEglCreateContext = NULL;
PFNPROXYEGLDESTROYCONTEXT                pfnProxyEglDestroyContext = NULL;
PFNPROXYEGLMAKECURRENT                   pfnProxyEglMakeCurrent = NULL;
PFNPROXYEGLGETCURRENTCONTEXT             pfnProxyEglGetCurrentContext = NULL;
PFNPROXYEGLGETCURRENTSURFACE             pfnProxyEglGetCurrentSurface = NULL;
PFNPROXYEGLGETCURRENTDISPLAY             pfnProxyEglGetCurrentDisplay = NULL;
PFNPROXYEGLQUERYCONTEXT                  pfnProxyEglQueryContext = NULL;
PFNPROXYEGLWAITGL                        pfnProxyEglWaitGL = NULL;
PFNPROXYEGLWAITNATIVE                    pfnProxyEglWaitNative = NULL;
PFNPROXYEGLSWAPBUFFERS                   pfnProxyEglSwapBuffers = NULL;
PFNPROXYEGLCOPYBUFFERS                   pfnProxyEglCopyBuffers = NULL;
PFNPROXYEGLGETPROCADDRESS                pfnProxyEglGetProcAddress = NULL;

extern void *glProxyGetProcAddr(void *handle, const char *name);

int eglProxyInit(void *handle)
{
    if (NULL == handle) {
        return -1;
    }

    if (!pfnProxyEglGetProcAddress)
        pfnProxyEglGetProcAddress = glProxyGetProcAddr(handle, "eglGetProcAddress");

    pfnProxyEglGetError = glProxyGetProcAddr(handle, "eglGetError");
    pfnProxyEglGetDisplay = glProxyGetProcAddr(handle, "eglGetDisplay");
    pfnProxyEglInitialize = glProxyGetProcAddr(handle, "eglInitialize");
    pfnProxyEglTerminate = glProxyGetProcAddr(handle, "eglTerminate");
    pfnProxyEglQueryString = glProxyGetProcAddr(handle, "eglQueryString");
    pfnProxyEglGetConfigs = glProxyGetProcAddr(handle, "eglGetConfigs");
    pfnProxyEglChooseConfig = glProxyGetProcAddr(handle, "eglChooseConfig");
    pfnProxyEglGetConfigAttrib = glProxyGetProcAddr(handle, "eglGetConfigAttrib");
    pfnProxyEglCreateWindowSurface = glProxyGetProcAddr(handle, "eglCreateWindowSurface");
    pfnProxyEglCreatePbufferSurface = glProxyGetProcAddr(handle, "eglCreatePbufferSurface");
    pfnProxyEglCreatePixmapSurface = glProxyGetProcAddr(handle, "eglCreatePixmapSurface");
    pfnProxyEglDestroySurface = glProxyGetProcAddr(handle, "eglDestroySurface");
    pfnProxyEglQuerySurface = glProxyGetProcAddr(handle, "eglQuerySurface");
    pfnProxyEglBindAPI = glProxyGetProcAddr(handle, "eglBindAPI");
    pfnProxyEglQueryAPI = glProxyGetProcAddr(handle, "eglQueryAPI");
    pfnProxyEglWaitClient = glProxyGetProcAddr(handle, "eglWaitClient");
    pfnProxyEglReleaseThread = glProxyGetProcAddr(handle, "eglReleaseThread");
    pfnProxyEglCreatePbufferFromClientBuffer = glProxyGetProcAddr(handle, "eglCreatePbufferFromClientBuffer");
    pfnProxyEglSurfaceAttrib = glProxyGetProcAddr(handle, "eglSurfaceAttrib");
    pfnProxyEglBindTexImage = glProxyGetProcAddr(handle, "eglBindTexImage");
    pfnProxyEglReleaseTexImage = glProxyGetProcAddr(handle, "eglReleaseTexImage");
    pfnProxyEglSwapInterval = glProxyGetProcAddr(handle, "eglSwapInterval");
    pfnProxyEglCreateContext = glProxyGetProcAddr(handle, "eglCreateContext");
    pfnProxyEglDestroyContext = glProxyGetProcAddr(handle, "eglDestroyContext");
    pfnProxyEglMakeCurrent = glProxyGetProcAddr(handle, "eglMakeCurrent");
    pfnProxyEglGetCurrentContext = glProxyGetProcAddr(handle, "eglGetCurrentContext");
    pfnProxyEglGetCurrentSurface = glProxyGetProcAddr(handle, "eglGetCurrentSurface");
    pfnProxyEglGetCurrentDisplay = glProxyGetProcAddr(handle, "eglGetCurrentDisplay");
    pfnProxyEglQueryContext = glProxyGetProcAddr(handle, "eglQueryContext");
    pfnProxyEglWaitGL = glProxyGetProcAddr(handle, "eglWaitGL");
    pfnProxyEglWaitNative = glProxyGetProcAddr(handle, "eglWaitNative");
    pfnProxyEglSwapBuffers = glProxyGetProcAddr(handle, "eglSwapBuffers");
    pfnProxyEglCopyBuffers = glProxyGetProcAddr(handle, "eglCopyBuffers");

    return 0;
}
