/* Copyright (C) 2010-2011 Linaro Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/*
 * Authors:
 *  Jammy Zhou <jammy.zhou@linaro.org>
 *  Marc Ordinas i Llopis, Collabora Ltd <marc.ordinasillopis@linaro.org>
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <string.h>
#include "GL/gl_proxy.h"
#include "GL/glx_proxy.h"
#include "EGL/egl_proxy.h"

/* global proxy context */
PROXY_CONTEXT global_proxy_context;

/* library names for different backends */
const char *opengl_libraries[] = {
    "libGL.so.1",
    "libGLESv2.so.2",
};

#define LIBEGL_PATH     "libEGL.so.1"

/* proxy functions */
PFNPROXYACTIVETEXTUREPROC pfnProxyActiveTexture = NULL;
PFNPROXYATTACHSHADERPROC pfnProxyAttachShader = NULL;
PFNPROXYBINDATTRIBLOCATIONPROC pfnProxyBindAttribLocation = NULL;
PFNPROXYBINDBUFFERPROC pfnProxyBindBuffer = NULL;
PFNPROXYBINDFRAMEBUFFERPROC pfnProxyBindFramebuffer = NULL;
PFNPROXYBINDRENDERBUFFERPROC pfnProxyBindRenderbuffer = NULL;
PFNPROXYBINDTEXTUREPROC pfnProxyBindTexture = NULL;
PFNPROXYBLENDCOLORPROC pfnProxyBlendColor = NULL;
PFNPROXYBLENDEQUATIONPROC pfnProxyBlendEquation = NULL;
PFNPROXYBLENDEQUATIONSEPARATEPROC pfnProxyBlendEquationSeparate = NULL;
PFNPROXYBLENDFUNCPROC pfnProxyBlendFunc = NULL;
PFNPROXYBLENDFUNCSEPARATEPROC pfnProxyBlendFuncSeparate = NULL;
PFNPROXYBUFFERDATAPROC pfnProxyBufferData = NULL;
PFNPROXYBUFFERSUBDATAPROC pfnProxyBufferSubData = NULL;
PFNPROXYCHECKFRAMEBUFFERSTATUSPROC pfnProxyCheckFramebufferStatus = NULL;
PFNPROXYCLEARPROC pfnProxyClear = NULL;
PFNPROXYCLEARCOLORPROC pfnProxyClearColor = NULL;
PFNPROXYCLEARDEPTHFPROC pfnProxyClearDepthf = NULL;
PFNPROXYCLEARSTENCILPROC pfnProxyClearStencil = NULL;
PFNPROXYCOLORMASKPROC pfnProxyColorMask = NULL;
PFNPROXYCOMPILESHADERPROC pfnProxyCompileShader = NULL;
PFNPROXYCOMPRESSEDTEXIMAGE2DPROC pfnProxyCompressedTexImage2D = NULL;
PFNPROXYCOMPRESSEDTEXSUBIMAGE2DPROC pfnProxyCompressedTexSubImage2D = NULL;
PFNPROXYCOPYTEXIMAGE2DPROC pfnProxyCopyTexImage2D = NULL;
PFNPROXYCOPYTEXSUBIMAGE2DPROC pfnProxyCopyTexSubImage2D = NULL;
PFNPROXYCREATEPROGRAMPROC pfnProxyCreateProgram = NULL;
PFNPROXYCREATESHADERPROC pfnProxyCreateShader = NULL;
PFNPROXYCULLFACEPROC pfnProxyCullFace = NULL;
PFNPROXYDELETEBUFFERSPROC pfnProxyDeleteBuffers = NULL;
PFNPROXYDELETEFRAMEBUFFERSPROC pfnProxyDeleteFramebuffers = NULL;
PFNPROXYDELETEPROGRAMPROC pfnProxyDeleteProgram = NULL;
PFNPROXYDELETERENDERBUFFERSPROC pfnProxyDeleteRenderbuffers = NULL;
PFNPROXYDELETESHADERPROC pfnProxyDeleteShader = NULL;
PFNPROXYDELETETEXTURESPROC pfnProxyDeleteTextures = NULL;
PFNPROXYDEPTHFUNCPROC pfnProxyDepthFunc = NULL;
PFNPROXYDEPTHMASKPROC pfnProxyDepthMask = NULL;
PFNPROXYDEPTHRANGEFPROC pfnProxyDepthRangef = NULL;
PFNPROXYDETACHSHADERPROC pfnProxyDetachShader = NULL;
PFNPROXYDISABLEPROC pfnProxyDisable = NULL;
PFNPROXYDISABLEVERTEXATTRIBARRAYPROC pfnProxyDisableVertexAttribArray = NULL;
PFNPROXYDRAWARRAYSPROC pfnProxyDrawArrays = NULL;
PFNPROXYDRAWELEMENTSPROC pfnProxyDrawElements = NULL;
PFNPROXYENABLEPROC pfnProxyEnable = NULL;
PFNPROXYENABLEVERTEXATTRIBARRAYPROC pfnProxyEnableVertexAttribArray = NULL;
PFNPROXYFINISHPROC pfnProxyFinish = NULL;
PFNPROXYFLUSHPROC pfnProxyFlush = NULL;
PFNPROXYFRAMEBUFFERRENDERBUFFERPROC pfnProxyFramebufferRenderbuffer = NULL;
PFNPROXYFRAMEBUFFERTEXTURE2DPROC pfnProxyFramebufferTexture2D = NULL;
PFNPROXYFRONTFACEPROC pfnProxyFrontFace = NULL;
PFNPROXYGENBUFFERSPROC pfnProxyGenBuffers = NULL;
PFNPROXYGENERATEMIPMAPPROC pfnProxyGenerateMipmap = NULL;
PFNPROXYGENFRAMEBUFFERSPROC pfnProxyGenFramebuffers = NULL;
PFNPROXYGENRENDERBUFFERSPROC pfnProxyGenRenderbuffers = NULL;
PFNPROXYGENTEXTURESPROC pfnProxyGenTextures = NULL;
PFNPROXYGETACTIVEATTRIBPROC pfnProxyGetActiveAttrib = NULL;
PFNPROXYGETACTIVEUNIFORMPROC pfnProxyGetActiveUniform = NULL;
PFNPROXYGETATTACHEDSHADERSPROC pfnProxyGetAttachedShaders = NULL;
PFNPROXYGETATTRIBLOCATIONPROC pfnProxyGetAttribLocation = NULL;
PFNPROXYGETBOOLEANVPROC pfnProxyGetBooleanv = NULL;
PFNPROXYGETBUFFERPARAMETERIVPROC pfnProxyGetBufferParameteriv = NULL;
PFNPROXYGETERRORPROC pfnProxyGetError = NULL;
PFNPROXYGETFLOATVPROC pfnProxyGetFloatv = NULL;
PFNPROXYGETFRAMEBUFFERATTACHMENTPARAMETERIVPROC pfnProxyGetFramebufferAttachmentParameteriv = NULL;
PFNPROXYGETINTEGERVPROC pfnProxyGetIntegerv = NULL;
PFNPROXYGETPROGRAMIVPROC pfnProxyGetProgramiv = NULL;
PFNPROXYGETPROGRAMINFOLOGPROC pfnProxyGetProgramInfoLog = NULL;
PFNPROXYGETRENDERBUFFERPARAMETERIVPROC pfnProxyGetRenderbufferParameteriv = NULL;
PFNPROXYGETSHADERIVPROC pfnProxyGetShaderiv = NULL;
PFNPROXYGETSHADERINFOLOGPROC pfnProxyGetShaderInfoLog = NULL;
PFNPROXYGETSHADERPRECISIONFORMATPROC pfnProxyGetShaderPrecisionFormat = NULL;
PFNPROXYGETSHADERSOURCEPROC pfnProxyGetShaderSource = NULL;
PFNPROXYGETSTRINGPROC pfnProxyGetString = NULL;
PFNPROXYGETTEXPARAMETERFVPROC pfnProxyGetTexParameterfv = NULL;
PFNPROXYGETTEXPARAMETERIVPROC pfnProxyGetTexParameteriv = NULL;
PFNPROXYGETUNIFORMFVPROC pfnProxyGetUniformfv = NULL;
PFNPROXYGETUNIFORMIVPROC pfnProxyGetUniformiv = NULL;
PFNPROXYGETUNIFORMLOCATIONPROC pfnProxyGetUniformLocation = NULL;
PFNPROXYGETVERTEXATTRIBFVPROC pfnProxyGetVertexAttribfv = NULL;
PFNPROXYGETVERTEXATTRIBIVPROC pfnProxyGetVertexAttribiv = NULL;
PFNPROXYGETVERTEXATTRIBPOINTERVPROC pfnProxyGetVertexAttribPointerv = NULL;
PFNPROXYHINTPROC pfnProxyHint = NULL;
PFNPROXYISBUFFERPROC pfnProxyIsBuffer = NULL;
PFNPROXYISENABLEDPROC pfnProxyIsEnabled = NULL;
PFNPROXYISFRAMEBUFFERPROC pfnProxyIsFramebuffer = NULL;
PFNPROXYISPROGRAMPROC pfnProxyIsProgram = NULL;
PFNPROXYISRENDERBUFFERPROC pfnProxyIsRenderbuffer = NULL;
PFNPROXYISSHADERPROC pfnProxyIsShader = NULL;
PFNPROXYISTEXTUREPROC pfnProxyIsTexture = NULL;
PFNPROXYLINEWIDTHPROC pfnProxyLineWidth = NULL;
PFNPROXYLINKPROGRAMPROC pfnProxyLinkProgram = NULL;
PFNPROXYPIXELSTOREIPROC pfnProxyPixelStorei = NULL;
PFNPROXYPOLYGONOFFSETPROC pfnProxyPolygonOffset = NULL;
PFNPROXYREADPIXELSPROC pfnProxyReadPixels = NULL;
PFNPROXYRELEASESHADERCOMPILERPROC pfnProxyReleaseShaderCompiler = NULL;
PFNPROXYRENDERBUFFERSTORAGEPROC pfnProxyRenderbufferStorage = NULL;
PFNPROXYSAMPLECOVERAGEPROC pfnProxySampleCoverage = NULL;
PFNPROXYSCISSORPROC pfnProxyScissor = NULL;
PFNPROXYSHADERBINARYPROC pfnProxyShaderBinary = NULL;
PFNPROXYSHADERSOURCEPROC pfnProxyShaderSource = NULL;
PFNPROXYSTENCILFUNCPROC pfnProxyStencilFunc = NULL;
PFNPROXYSTENCILFUNCSEPARATEPROC pfnProxyStencilFuncSeparate = NULL;
PFNPROXYSTENCILMASKPROC pfnProxyStencilMask = NULL;
PFNPROXYSTENCILMASKSEPARATEPROC pfnProxyStencilMaskSeparate = NULL;
PFNPROXYSTENCILOPPROC pfnProxyStencilOp = NULL;
PFNPROXYSTENCILOPSEPARATEPROC pfnProxyStencilOpSeparate = NULL;
PFNPROXYTEXIMAGE2DPROC pfnProxyTexImage2D = NULL;
PFNPROXYTEXPARAMETERFPROC pfnProxyTexParameterf = NULL;
PFNPROXYTEXPARAMETERFVPROC pfnProxyTexParameterfv = NULL;
PFNPROXYTEXPARAMETERIPROC pfnProxyTexParameteri = NULL;
PFNPROXYTEXPARAMETERIVPROC pfnProxyTexParameteriv = NULL;
PFNPROXYTEXSUBIMAGE2DPROC pfnProxyTexSubImage2D = NULL;
PFNPROXYUNIFORM1FPROC pfnProxyUniform1f = NULL;
PFNPROXYUNIFORM1FVPROC pfnProxyUniform1fv = NULL;
PFNPROXYUNIFORM1IPROC pfnProxyUniform1i = NULL;
PFNPROXYUNIFORM1IVPROC pfnProxyUniform1iv = NULL;
PFNPROXYUNIFORM2FPROC pfnProxyUniform2f = NULL;
PFNPROXYUNIFORM2FVPROC pfnProxyUniform2fv = NULL;
PFNPROXYUNIFORM2IPROC pfnProxyUniform2i = NULL;
PFNPROXYUNIFORM2IVPROC pfnProxyUniform2iv = NULL;
PFNPROXYUNIFORM3FPROC pfnProxyUniform3f = NULL;
PFNPROXYUNIFORM3FVPROC pfnProxyUniform3fv = NULL;
PFNPROXYUNIFORM3IPROC pfnProxyUniform3i = NULL;
PFNPROXYUNIFORM3IVPROC pfnProxyUniform3iv = NULL;
PFNPROXYUNIFORM4FPROC pfnProxyUniform4f = NULL;
PFNPROXYUNIFORM4FVPROC pfnProxyUniform4fv = NULL;
PFNPROXYUNIFORM4IPROC pfnProxyUniform4i = NULL;
PFNPROXYUNIFORM4IVPROC pfnProxyUniform4iv = NULL;
PFNPROXYUNIFORMMATRIX2FVPROC pfnProxyUniformMatrix2fv = NULL;
PFNPROXYUNIFORMMATRIX3FVPROC pfnProxyUniformMatrix3fv = NULL;
PFNPROXYUNIFORMMATRIX4FVPROC pfnProxyUniformMatrix4fv = NULL;
PFNPROXYUSEPROGRAMPROC pfnProxyUseProgram = NULL;
PFNPROXYVALIDATEPROGRAMPROC pfnProxyValidateProgram = NULL;
PFNPROXYVERTEXATTRIB1FPROC pfnProxyVertexAttrib1f = NULL;
PFNPROXYVERTEXATTRIB1FVPROC pfnProxyVertexAttrib1fv = NULL;
PFNPROXYVERTEXATTRIB2FPROC pfnProxyVertexAttrib2f = NULL;
PFNPROXYVERTEXATTRIB2FVPROC pfnProxyVertexAttrib2fv = NULL;
PFNPROXYVERTEXATTRIB3FPROC pfnProxyVertexAttrib3f = NULL;
PFNPROXYVERTEXATTRIB3FVPROC pfnProxyVertexAttrib3fv = NULL;
PFNPROXYVERTEXATTRIB4FPROC pfnProxyVertexAttrib4f = NULL;
PFNPROXYVERTEXATTRIB4FVPROC pfnProxyVertexAttrib4fv = NULL;
PFNPROXYVERTEXATTRIBPOINTERPROC pfnProxyVertexAttribPointer = NULL;
PFNPROXYVIEWPORTPROC pfnProxyViewport = NULL;

static void *glProxyOpenLibrary(const char *name)
{
    return dlopen(name, RTLD_LAZY | RTLD_GLOBAL);
}

static int glProxyCloseLibrary(void *handle)
{
    return dlclose(handle);
}

static int glProxyEarlyResolve(PROXY_CONTEXT *proxy_context) {
    if (OPENGL_BACKEND == proxy_context->current_backend) {
        proxy_context->gl_handle = glProxyOpenLibrary(opengl_libraries[OPENGL_BACKEND]);

        pfnProxyGLXGetProcAddress = dlsym(proxy_context->gl_handle, "glXGetProcAddress");
        if (NULL == pfnProxyGLXGetProcAddress) {
            pfnProxyGLXGetProcAddressARB = dlsym(proxy_context->gl_handle,
                    "glXGetProcAddressARB");
        }
    } else if (OPENGL_ES20_BACKEND == proxy_context->current_backend) {
        proxy_context->egl_handle = glProxyOpenLibrary(LIBEGL_PATH);
        pfnProxyEglGetProcAddress = dlsym(proxy_context->egl_handle, "eglGetProcAddress");
    }

    return 0;
}

void *glProxyGetProcAddr(void *handle, const char *name) {
    if (pfnProxyGLXGetProcAddress) {
        return (*pfnProxyGLXGetProcAddress)((const GLubyte *) name);
    } else if (pfnProxyGLXGetProcAddressARB) {
        return (*pfnProxyGLXGetProcAddressARB)((const GLubyte *) name);
    } else {
        return dlsym(handle, name);
    }
}

static GLboolean glProxyCheckExtension(const char *list, const char *name)
{
    size_t len = strlen(name);
    const char *ext_ptr = list;

    if(ext_ptr == NULL)
        return GL_FALSE;

    while ((ext_ptr = strstr(ext_ptr, name)) != NULL) {
        if (ext_ptr[len] == ' ' || ext_ptr[len] == '\0')
            break;
        ext_ptr += len;
    }

    return (ext_ptr != NULL)? GL_TRUE : GL_FALSE;
}

GLboolean glProxyHasExtension(void *display, const char *extension)
{
    char const *backend_extensions = NULL;
    char const *gl_extensions = NULL;

    if (global_proxy_context.current_backend == UNKNOWN_BACKEND)
        return GL_FALSE;

    /* Get the backend extensions string */
    if (global_proxy_context.current_backend == OPENGL_BACKEND &&
        pfnProxyQueryExtensionsString) {
        backend_extensions = pfnProxyQueryExtensionsString(
                                 display,
                                 DefaultScreen(display));
    } else if (global_proxy_context.current_backend == OPENGL_ES20_BACKEND &&
               pfnProxyEglQueryString) {
        backend_extensions = pfnProxyEglQueryString(
                                (EGLDisplay)display,
                                EGL_EXTENSIONS);
    }

    /* Get the GL extensions string */
    if (pfnProxyGetString)
        gl_extensions = (const char *)pfnProxyGetString(GL_EXTENSIONS);

    /* Check if present */
    if (glProxyCheckExtension(backend_extensions, extension) ||
        glProxyCheckExtension(gl_extensions, extension))
        return GL_TRUE;

    return GL_FALSE;
}

void *glProxyGetExtProcAddr(void *display, const char *extension, const char *name) {
    if (!glProxyHasExtension(display, extension))
        return NULL;

    if (pfnProxyEglGetProcAddress) {
        return (*pfnProxyEglGetProcAddress)(name);
    } else if (pfnProxyGLXGetProcAddress) {
        return (*pfnProxyGLXGetProcAddress)((const GLubyte *) name);
    } else if (pfnProxyGLXGetProcAddressARB) {
        return (*pfnProxyGLXGetProcAddressARB)((const GLubyte *) name);
    }

    return NULL;
}

static void glProxyBackendSelection(PROXY_BACKEND_OPTION *pProxyBackend, PROXY_CONTEXT *proxy_context)
{
    PROXY_BACKEND_OPTION option = *pProxyBackend;

    if (OPENGL_BACKEND == option) {
        if ((proxy_context->gl_handle = glProxyOpenLibrary(
                opengl_libraries[OPENGL_BACKEND]))) {
            *pProxyBackend = OPENGL_BACKEND;
        } else if ((proxy_context->gles2_handle = glProxyOpenLibrary(
                opengl_libraries[OPENGL_ES20_BACKEND]))) {
            *pProxyBackend = OPENGL_ES20_BACKEND;
        } else {
            *pProxyBackend = UNKNOWN_BACKEND;
        }
    } else if (OPENGL_ES20_BACKEND == option) {
        if ((proxy_context->gles2_handle = glProxyOpenLibrary(
                opengl_libraries[OPENGL_ES20_BACKEND]))) {
            *pProxyBackend = OPENGL_ES20_BACKEND;
        } else if ((proxy_context->gl_handle = glProxyOpenLibrary(
                opengl_libraries[OPENGL_BACKEND]))) {
            *pProxyBackend = OPENGL_BACKEND;
        } else {
            *pProxyBackend = UNKNOWN_BACKEND;
        }
    } else {
        /* Select backend automatically */
        proxy_context->gles2_handle = glProxyOpenLibrary(
                opengl_libraries[OPENGL_ES20_BACKEND]);
        proxy_context->egl_handle = glProxyOpenLibrary(LIBEGL_PATH);
        proxy_context->gl_handle = glProxyOpenLibrary(
                opengl_libraries[OPENGL_BACKEND]);

        if (proxy_context->gles2_handle) {
            /* Use GLES2 by default if available */
            *pProxyBackend = OPENGL_ES20_BACKEND;
        } else if (proxy_context->gl_handle) {
            *pProxyBackend = OPENGL_BACKEND;
        } else {
            *pProxyBackend = UNKNOWN_BACKEND;
        }

        if (OPENGL_ES20_BACKEND == *pProxyBackend) {
            if (proxy_context->gl_handle) {
                glProxyCloseLibrary(proxy_context->gl_handle);
                proxy_context->gl_handle = NULL;
            }
        } else if (OPENGL_BACKEND == *pProxyBackend) {
            if (proxy_context->egl_handle) {
                glProxyCloseLibrary(proxy_context->egl_handle);
                proxy_context->egl_handle = NULL;
            }

            if (proxy_context->gles2_handle) {
                glProxyCloseLibrary(proxy_context->gles2_handle);
                proxy_context->gles2_handle = NULL;
            }
        }
    }

    proxy_context->current_backend = *pProxyBackend;
}

static int glProxyCoreInit(void *handle)
{
	if (NULL == handle) {
        return -1;
    }

    pfnProxyActiveTexture = glProxyGetProcAddr(handle, "glActiveTexture");
    pfnProxyAttachShader = glProxyGetProcAddr(handle, "glAttachShader");
    pfnProxyBindAttribLocation = glProxyGetProcAddr(handle,
            "glBindAttribLocation");
    pfnProxyBindBuffer = glProxyGetProcAddr(handle, "glBindBuffer");
    pfnProxyBindFramebuffer = glProxyGetProcAddr(handle, "glBindFramebuffer");
    pfnProxyBindRenderbuffer = glProxyGetProcAddr(handle, "glBindRenderbuffer");
    pfnProxyBindTexture = glProxyGetProcAddr(handle, "glBindTexture");
    pfnProxyBlendColor = glProxyGetProcAddr(handle, "glBlendColor");
    pfnProxyBlendEquation = glProxyGetProcAddr(handle, "glBlendEquation");
    pfnProxyBlendEquationSeparate = glProxyGetProcAddr(handle,
            "glBlendEquationSeparate");
    pfnProxyBlendFunc = glProxyGetProcAddr(handle, "glBlendFunc");
    pfnProxyBlendFuncSeparate = glProxyGetProcAddr(handle,
            "glBlendFuncSeparate");
    pfnProxyBufferData = glProxyGetProcAddr(handle, "glBufferData");
    pfnProxyBufferSubData = glProxyGetProcAddr(handle, "glBufferSubData");
    pfnProxyCheckFramebufferStatus = glProxyGetProcAddr(handle,
            "glCheckFramebufferStatus");
    pfnProxyClear = glProxyGetProcAddr(handle, "glClear");
    pfnProxyClearColor = glProxyGetProcAddr(handle, "glClearColor");
    pfnProxyClearDepthf = glProxyGetProcAddr(handle, "glClearDepthf");
    pfnProxyClearStencil = glProxyGetProcAddr(handle, "glClearStencil");
    pfnProxyColorMask = glProxyGetProcAddr(handle, "glColorMask");
    pfnProxyCompileShader = glProxyGetProcAddr(handle, "glCompileShader");
    pfnProxyCompressedTexImage2D = glProxyGetProcAddr(handle,
            "glCompressedTexImage2D");
    pfnProxyCompressedTexSubImage2D = glProxyGetProcAddr(handle,
            "glCompressedTexSubImage2D");
    pfnProxyCopyTexImage2D = glProxyGetProcAddr(handle, "glCopyTexImage2D");
    pfnProxyCopyTexSubImage2D = glProxyGetProcAddr(handle,
            "glCopyTexSubImage2D");
    pfnProxyCreateProgram = glProxyGetProcAddr(handle, "glCreateProgram");
    pfnProxyCreateShader = glProxyGetProcAddr(handle, "glCreateShader");
    pfnProxyCullFace = glProxyGetProcAddr(handle, "glCullFace");
    pfnProxyDeleteBuffers = glProxyGetProcAddr(handle, "glDeleteBuffers");
    pfnProxyDeleteFramebuffers = glProxyGetProcAddr(handle,
            "glDeleteFramebuffers");
    pfnProxyDeleteProgram = glProxyGetProcAddr(handle, "glDeleteProgram");
    pfnProxyDeleteRenderbuffers = glProxyGetProcAddr(handle,
            "glDeleteRenderbuffers");
    pfnProxyDeleteShader = glProxyGetProcAddr(handle, "glDeleteShader");
    pfnProxyDeleteTextures = glProxyGetProcAddr(handle, "glDeleteTextures");
    pfnProxyDepthFunc = glProxyGetProcAddr(handle, "glDepthFunc");
    pfnProxyDepthMask = glProxyGetProcAddr(handle, "glDepthMask");
    pfnProxyDepthRangef = glProxyGetProcAddr(handle, "glDepthRangef");
    pfnProxyDetachShader = glProxyGetProcAddr(handle, "glDetachShader");
    pfnProxyDisable = glProxyGetProcAddr(handle, "glDisable");
    pfnProxyDisableVertexAttribArray = glProxyGetProcAddr(handle,
            "glDisableVertexAttribArray");
    pfnProxyDrawArrays = glProxyGetProcAddr(handle, "glDrawArrays");
    pfnProxyDrawElements = glProxyGetProcAddr(handle, "glDrawElements");
    pfnProxyEnable = glProxyGetProcAddr(handle, "glEnable");
    pfnProxyEnableVertexAttribArray = glProxyGetProcAddr(handle,
            "glEnableVertexAttribArray");
    pfnProxyFinish = glProxyGetProcAddr(handle, "glFinish");
    pfnProxyFlush = glProxyGetProcAddr(handle, "glFlush");
    pfnProxyFramebufferRenderbuffer = glProxyGetProcAddr(handle,
            "glFramebufferRenderbuffer");
    pfnProxyFramebufferTexture2D = glProxyGetProcAddr(handle,
            "glFramebufferTexture2D");
    pfnProxyFrontFace = glProxyGetProcAddr(handle, "glFrontFace");
    pfnProxyGenBuffers = glProxyGetProcAddr(handle, "glGenBuffers");
    pfnProxyGenerateMipmap = glProxyGetProcAddr(handle, "glGenerateMipmap");
    pfnProxyGenFramebuffers = glProxyGetProcAddr(handle, "glGenFramebuffers");
    pfnProxyGenRenderbuffers = glProxyGetProcAddr(handle, "glGenRenderbuffers");
    pfnProxyGenTextures = glProxyGetProcAddr(handle, "glGenTextures");
    pfnProxyGetActiveAttrib = glProxyGetProcAddr(handle, "glGetActiveAttrib");
    pfnProxyGetActiveUniform = glProxyGetProcAddr(handle, "glGetActiveUniform");
    pfnProxyGetAttachedShaders = glProxyGetProcAddr(handle,
            "glGetAttachedShaders");
    pfnProxyGetAttribLocation = glProxyGetProcAddr(handle,
            "glGetAttribLocation");
    pfnProxyGetBooleanv = glProxyGetProcAddr(handle, "glGetBooleanv");
    pfnProxyGetBufferParameteriv = glProxyGetProcAddr(handle,
            "glGetBufferParameteriv");
    pfnProxyGetError = glProxyGetProcAddr(handle, "glGetError");
    pfnProxyGetFloatv = glProxyGetProcAddr(handle, "glGetFloatv");
    pfnProxyGetFramebufferAttachmentParameteriv = glProxyGetProcAddr(handle,
            "glGetFramebufferAttachmentParameteriv");
    pfnProxyGetIntegerv = glProxyGetProcAddr(handle, "glGetIntegerv");
    pfnProxyGetProgramiv = glProxyGetProcAddr(handle, "glGetProgramiv");
    pfnProxyGetProgramInfoLog = glProxyGetProcAddr(handle,
            "glGetProgramInfoLog");
    pfnProxyGetRenderbufferParameteriv = glProxyGetProcAddr(handle,
            "glGetRenderbufferParameteriv");
    pfnProxyGetShaderiv = glProxyGetProcAddr(handle, "glGetShaderiv");
    pfnProxyGetShaderInfoLog = glProxyGetProcAddr(handle, "glGetShaderInfoLog");
    pfnProxyGetShaderPrecisionFormat = glProxyGetProcAddr(handle,
            "glGetShaderPrecisionFormat");
    pfnProxyGetShaderSource = glProxyGetProcAddr(handle, "glGetShaderSource");
    pfnProxyGetString = glProxyGetProcAddr(handle, "glGetString");
    pfnProxyGetTexParameterfv = glProxyGetProcAddr(handle,
            "glGetTexParameterfv");
    pfnProxyGetTexParameteriv = glProxyGetProcAddr(handle,
            "glGetTexParameteriv");
    pfnProxyGetUniformfv = glProxyGetProcAddr(handle, "glGetUniformfv");
    pfnProxyGetUniformiv = glProxyGetProcAddr(handle, "glGetUniformiv");
    pfnProxyGetUniformLocation = glProxyGetProcAddr(handle,
            "glGetUniformLocation");
    pfnProxyGetVertexAttribfv = glProxyGetProcAddr(handle,
            "glGetVertexAttribfv");
    pfnProxyGetVertexAttribiv = glProxyGetProcAddr(handle,
            "glGetVertexAttribiv");
    pfnProxyGetVertexAttribPointerv = glProxyGetProcAddr(handle,
            "glGetVertexAttribPointerv");
    pfnProxyHint = glProxyGetProcAddr(handle, "glHint");
    pfnProxyIsBuffer = glProxyGetProcAddr(handle, "glIsBuffer");
    pfnProxyIsEnabled = glProxyGetProcAddr(handle, "glIsEnabled");
    pfnProxyIsFramebuffer = glProxyGetProcAddr(handle, "glIsFramebuffer");
    pfnProxyIsProgram = glProxyGetProcAddr(handle, "glIsProgram");
    pfnProxyIsRenderbuffer = glProxyGetProcAddr(handle, "glIsRenderbuffer");
    pfnProxyIsShader = glProxyGetProcAddr(handle, "glIsShader");
    pfnProxyIsTexture = glProxyGetProcAddr(handle, "glIsTexture");
    pfnProxyLineWidth = glProxyGetProcAddr(handle, "glLineWidth");
    pfnProxyLinkProgram = glProxyGetProcAddr(handle, "glLinkProgram");
    pfnProxyPixelStorei = glProxyGetProcAddr(handle, "glPixelStorei");
    pfnProxyPolygonOffset = glProxyGetProcAddr(handle, "glPolygonOffset");
    pfnProxyReadPixels = glProxyGetProcAddr(handle, "glReadPixels");
    pfnProxyReleaseShaderCompiler = glProxyGetProcAddr(handle,
            "glReleaseShaderCompiler");
    pfnProxyRenderbufferStorage = glProxyGetProcAddr(handle,
            "glRenderbufferStorage");
    pfnProxySampleCoverage = glProxyGetProcAddr(handle, "glSampleCoverage");
    pfnProxyScissor = glProxyGetProcAddr(handle, "glScissor");
    pfnProxyShaderBinary = glProxyGetProcAddr(handle, "glShaderBinary");
    pfnProxyShaderSource = glProxyGetProcAddr(handle, "glShaderSource");
    pfnProxyStencilFunc = glProxyGetProcAddr(handle, "glStencilFunc");
    pfnProxyStencilFuncSeparate = glProxyGetProcAddr(handle,
            "glStencilFuncSeparate");
    pfnProxyStencilMask = glProxyGetProcAddr(handle, "glStencilMask");
    pfnProxyStencilMaskSeparate = glProxyGetProcAddr(handle,
            "glStencilMaskSeparate");
    pfnProxyStencilOp = glProxyGetProcAddr(handle, "glStencilOp");
    pfnProxyStencilOpSeparate = glProxyGetProcAddr(handle,
            "glStencilOpSeparate");
    pfnProxyTexImage2D = glProxyGetProcAddr(handle, "glTexImage2D");
    pfnProxyTexParameterf = glProxyGetProcAddr(handle, "glTexParameterf");
    pfnProxyTexParameterfv = glProxyGetProcAddr(handle, "glTexParameterfv");
    pfnProxyTexParameteri = glProxyGetProcAddr(handle, "glTexParameteri");
    pfnProxyTexParameteriv = glProxyGetProcAddr(handle, "glTexParameteriv");
    pfnProxyTexSubImage2D = glProxyGetProcAddr(handle, "glTexSubImage2D");
    pfnProxyUniform1f = glProxyGetProcAddr(handle, "glUniform1f");
    pfnProxyUniform1fv = glProxyGetProcAddr(handle, "glUniform1fv");
    pfnProxyUniform1i = glProxyGetProcAddr(handle, "glUniform1i");
    pfnProxyUniform1iv = glProxyGetProcAddr(handle, "glUniform1iv");
    pfnProxyUniform2f = glProxyGetProcAddr(handle, "glUniform2f");
    pfnProxyUniform2fv = glProxyGetProcAddr(handle, "glUniform2fv");
    pfnProxyUniform2i = glProxyGetProcAddr(handle, "glUniform2i");
    pfnProxyUniform2iv = glProxyGetProcAddr(handle, "glUniform2iv");
    pfnProxyUniform3f = glProxyGetProcAddr(handle, "glUniform3f");
    pfnProxyUniform3fv = glProxyGetProcAddr(handle, "glUniform3fv");
    pfnProxyUniform3i = glProxyGetProcAddr(handle, "glUniform3i");
    pfnProxyUniform3iv = glProxyGetProcAddr(handle, "glUniform3iv");
    pfnProxyUniform4f = glProxyGetProcAddr(handle, "glUniform4f");
    pfnProxyUniform4fv = glProxyGetProcAddr(handle, "glUniform4fv");
    pfnProxyUniform4i = glProxyGetProcAddr(handle, "glUniform4i");
    pfnProxyUniform4iv = glProxyGetProcAddr(handle, "glUniform4iv");
    pfnProxyUniformMatrix2fv = glProxyGetProcAddr(handle, "glUniformMatrix2fv");
    pfnProxyUniformMatrix3fv = glProxyGetProcAddr(handle, "glUniformMatrix3fv");
    pfnProxyUniformMatrix4fv = glProxyGetProcAddr(handle, "glUniformMatrix4fv");
    pfnProxyUseProgram = glProxyGetProcAddr(handle, "glUseProgram");
    pfnProxyValidateProgram = glProxyGetProcAddr(handle, "glValidateProgram");
    pfnProxyVertexAttrib1f = glProxyGetProcAddr(handle, "glVertexAttrib1f");
    pfnProxyVertexAttrib1fv = glProxyGetProcAddr(handle, "glVertexAttrib1fv");
    pfnProxyVertexAttrib2f = glProxyGetProcAddr(handle, "glVertexAttrib2f");
    pfnProxyVertexAttrib2fv = glProxyGetProcAddr(handle, "glVertexAttrib2fv");
    pfnProxyVertexAttrib3f = glProxyGetProcAddr(handle, "glVertexAttrib3f");
    pfnProxyVertexAttrib3fv = glProxyGetProcAddr(handle, "glVertexAttrib3fv");
    pfnProxyVertexAttrib4f = glProxyGetProcAddr(handle, "glVertexAttrib4f");
    pfnProxyVertexAttrib4fv = glProxyGetProcAddr(handle, "glVertexAttrib4fv");
    pfnProxyVertexAttribPointer = glProxyGetProcAddr(handle,
            "glVertexAttribPointer");
    pfnProxyViewport = glProxyGetProcAddr(handle, "glViewport");

    return 0;
}

extern int glxProxyInit(void *handle);
extern int eglProxyInit(void *handle);

int glProxyInit(PROXY_BACKEND_OPTION *pProxyBackend)
{
    global_proxy_context.current_backend = UNKNOWN_BACKEND;

    /* Select backend first */
    glProxyBackendSelection(pProxyBackend, &global_proxy_context);

    if (UNKNOWN_BACKEND == global_proxy_context.current_backend) {
        printf("Error: No OpenGL or OpenGL ES2.0 renderer available\n");
        return -1;
    }

    printf("%s backend is selected for rendering\n",
            (global_proxy_context.current_backend == OPENGL_BACKEND) ? "OpenGL"
                    : "OpenGL ES2.0");

    /* Load selected backend, and resolve glXGetProcAddress/eglGetProcAddress */
    if (glProxyEarlyResolve(&global_proxy_context) < 0)
        return -1;

    /* Resolve symbols */
    if (OPENGL_BACKEND == global_proxy_context.current_backend) {
        glProxyCoreInit(global_proxy_context.gl_handle);
        glxProxyInit(global_proxy_context.gl_handle);
    } else if (OPENGL_ES20_BACKEND == global_proxy_context.current_backend) {
        glProxyCoreInit(global_proxy_context.gles2_handle);
        eglProxyInit(global_proxy_context.egl_handle);
    } else {
        /* should not get here */
        return -1;
    }

    return 0;
}

int glProxyClose(void)
{
    if (global_proxy_context.gl_handle) {
        glProxyCloseLibrary(global_proxy_context.gl_handle);
        global_proxy_context.gl_handle = NULL;
    }

    if (global_proxy_context.gles2_handle) {
        glProxyCloseLibrary(global_proxy_context.gles2_handle);
        global_proxy_context.gles2_handle = NULL;
    }

    if (global_proxy_context.egl_handle) {
        glProxyCloseLibrary(global_proxy_context.egl_handle);
        global_proxy_context.egl_handle = NULL;
    }

    global_proxy_context.current_backend = UNKNOWN_BACKEND;

    return 0;
}
